package com.landray.kmss.sys.attachment.util;

import java.security.NoSuchAlgorithmException;
import java.security.MessageDigest;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Collection;
import java.util.SortedMap;

public class ConsistentHash<T>
{
    private final HashAlgorithm hashAlgorithm;
    private final int numberOfReplicas;
    private final SortedMap<Long, T> circle;
    
    public ConsistentHash(final Collection<T> nodes) {
        this(160, nodes);
    }
    
    public ConsistentHash(final int numberOfReplicas, final Collection<T> nodes) {
        this.hashAlgorithm = new HashAlgorithm((HashAlgorithm)null);
        this.circle = new TreeMap<Long, T>();
        this.numberOfReplicas = numberOfReplicas;
        for (final T node : nodes) {
            this.add(node);
        }
    }
    
    public void add(final T node) {
        for (int i = 0; i < this.numberOfReplicas; ++i) {
            this.circle.put(this.hashAlgorithm.hash(String.valueOf(node.toString()) + i), node);
        }
    }
    
    public void remove(final T node) {
        for (int i = 0; i < this.numberOfReplicas; ++i) {
            this.circle.remove(this.hashAlgorithm.hash(String.valueOf(node.toString()) + i));
        }
    }
    
    public T get(final Object key) {
        if (this.circle.isEmpty()) {
            return null;
        }
        long hash = this.hashAlgorithm.hash((String)key);
        if (!this.circle.containsKey(hash)) {
            final SortedMap<Long, T> tailMap = this.circle.tailMap(hash);
            hash = (tailMap.isEmpty() ? this.circle.firstKey() : tailMap.firstKey());
        }
        return this.circle.get(hash);
    }
    
    public long getSize() {
        return this.circle.size();
    }
    
    private class HashAlgorithm
    {
        private MessageDigest md5;
        
        private HashAlgorithm() {
            this.md5 = null;
        }
        
        public long hash(final String key) {
            if (this.md5 == null) {
                try {
                    this.md5 = MessageDigest.getInstance("MD5");
                }
                catch (final NoSuchAlgorithmException ex) {
                    throw new IllegalStateException("no md5 algorythm found");
                }
            }
            this.md5.reset();
            this.md5.update(key.getBytes());
            final byte[] bKey = this.md5.digest();
            final long res = (long)(bKey[3] & 0xFF) << 24 | (long)(bKey[2] & 0xFF) << 16 | (long)(bKey[1] & 0xFF) << 8 | (long)(bKey[0] & 0xFF);
            return res & 0xFFFFFFFFL;
        }
    }
}
